import json
from typing import List
from io import BytesIO

import pandas as pd
from numpy import number, nan
from pandas.errors import ParserError
from werkzeug.datastructures import FileStorage

from file_process.base import FileProcessorBase
from file_process.exceptions import ModelFileValidationTargetsError, ModelFileValidationVariablesError, DelimiterError


class CSVFileProcessor(FileProcessorBase):
    def read_file(self, file, **kwargs):
        if isinstance(file, FileStorage):  # TODO try to get rid of it
            file = file.read()
            file = BytesIO(file)
        read_rows_count = kwargs.get('read_rows_count', 10)
        delimiter = kwargs.get('delimiter', None)
        data = self.read_csv_with_delimiter(file, read_rows_count, delimiter)
        return data

    def read_csv_with_delimiter(self, data_stream, read_rows_count: int, delimiter: str = None):
        if not delimiter:
            reader = pd.read_csv(data_stream, sep=None, iterator=True, nrows=read_rows_count)
            delimiter = reader._engine.data.dialect.delimiter  # pylint: disable=protected-access
            data_stream.seek(0)
        try:
            df = pd.read_csv(data_stream, sep=delimiter)
        except ParserError as exc:
            raise DelimiterError() from exc
        return df

    def get_preview_data(self, df: pd.DataFrame):
        var_names = list(df.columns)
        obs_preview = df.head(min(10, df.shape[0]))
        return var_names, obs_preview

    def model_file_validation(self, df: pd.DataFrame, model_metadata_file: BytesIO, need_target: bool = True):
        reader = json.load(model_metadata_file)
        var_names = set(reader['columns'])
        target_names = set(reader['targets'])
        metadata = reader.get('metadata', {})
        dataset_vars = set(df.columns)

        if need_target:
            all_targets = metadata.get('require_all_targets', True)
            if all_targets:
                are_targets_valid = not target_names or all(elem in dataset_vars for elem in target_names)
            else:
                are_targets_valid = not target_names or any(elem in dataset_vars for elem in target_names)
            if not are_targets_valid:
                raise ModelFileValidationTargetsError
        are_variables_valid = all(elem in dataset_vars.difference(target_names)
                                  for elem in var_names.difference(target_names))
        if not are_variables_valid:
            raise ModelFileValidationVariablesError

    def process(self, file, model_metadata_file: BytesIO = None, **kwargs) -> (List[str], None, pd.DataFrame):
        df = self.read_file(file, **kwargs)
        if model_metadata_file:
            self.model_file_validation(df, model_metadata_file)
        var_names, obs_preview = self.get_preview_data(df)
        return var_names, None, obs_preview

    def create_tabular_response(self, data_df: pd.DataFrame) -> List[dict]:
        if data_df is None:
            return []
        numeric_columns = data_df.select_dtypes(include=number).columns
        rows = data_df.replace({nan: None})
        rows[numeric_columns] = rows[numeric_columns].round(2)
        rows = rows.to_dict(orient='records')
        return rows



import anndata

from file_process.base import TabularFileProcessorBase


class H5ADFileProcessor(TabularFileProcessorBase):

    def read_file(self, file, **kwargs):
        return anndata.read_h5ad(file)



from typing import Optional, List

from file_process.file_processor_factory import FileProcessFactory


# pylint: disable=too-many-function-args
def preview_file(filename: str, file, *, model_metadata_file=None, **kwargs) \
        -> (List[str], Optional[List[dict]], Optional[List[dict]]):
    processor = FileProcessFactory.get(filename)
    target_names, var_preview, obs_preview = processor.process(file, model_metadata_file, **kwargs)
    return target_names, processor.create_tabular_response(var_preview), processor.create_tabular_response(obs_preview)



from file_process.csv import CSVFileProcessor
from file_process.exceptions import WrongExtension
from file_process.h5 import H5ADFileProcessor


class FileProcessFactory:  # pylint: disable=too-few-public-methods
    @classmethod
    def get(cls, filename: str):
        if filename.endswith('.h5ad'):
            return H5ADFileProcessor()
        if filename.endswith('.csv'):
            return CSVFileProcessor()
        raise WrongExtension



from abc import ABC, abstractmethod
from typing import List, Optional
from io import BytesIO

import pandas as pd
from anndata import AnnData
from pandas import notnull

from file_process.exceptions import ModelFileValidationError, NoColumnsError


class FileProcessorBase(ABC):
    @abstractmethod
    def read_file(self, file, **kwargs):
        raise NotImplemented

    @abstractmethod
    def process(self, file, model_metadata_file: BytesIO = None, **kwargs) \
            -> (List[str], Optional[pd.DataFrame], Optional[pd.DataFrame]):
        raise NotImplemented

    @abstractmethod
    def create_tabular_response(self, data_df: pd.DataFrame) -> List[dict]:
        raise NotImplemented


class TabularFileProcessorBase(FileProcessorBase, ABC):
    def get_preview_data(self, adata):
        target_names = list(adata.obs.columns)
        var_preview = adata.var.head(n=10)
        obs_preview = adata.obs.head(n=10)
        return target_names, var_preview, obs_preview

    def model_file_validation(self, adata: AnnData, model_metadata_file: BytesIO):
        reader = pd.read_csv(model_metadata_file, sep=',', index_col=0)
        var_names = reader.index
        dataset_vars = list(adata.var.index)
        result = all(elem in dataset_vars for elem in var_names)
        if not result:
            raise ModelFileValidationError

    def validate(self, adata: AnnData):
        target_names = list(adata.obs.columns)
        if not target_names:
            raise NoColumnsError

    def process(self, file, model_metadata_file: BytesIO = None, **kwargs) -> (List[str], pd.DataFrame, pd.DataFrame):
        adata = self.read_file(file, **kwargs)
        self.validate(adata)
        if model_metadata_file:
            self.model_file_validation(adata, model_metadata_file)
        target_names, var_preview, obs_preview = self.get_preview_data(adata)
        return target_names, var_preview, obs_preview

    def create_tabular_response(self, data_df: pd.DataFrame) -> List[dict]:
        if data_df is None:
            return []
        data_df = data_df.astype(object)
        rows = data_df.round(2).where(notnull(data_df), None).to_dict(orient='records')
        indices = list(data_df.index)
        for index, value in enumerate(indices):
            rows[index]['Feature Name'] = value
        return rows



from abc import ABC, abstractmethod


class BaseError(ABC, Exception):
    @property
    @abstractmethod
    def message(self):
        pass


class ModelFileValidationError(BaseError):
    message = 'Validation check failed: new data does not contain all fields required by model.'


class ModelFileValidationTargetsError(BaseError):
    message = 'Validation check failed: new data does not contain all targets required by model.'


class ModelFileValidationVariablesError(BaseError):
    message = 'Validation check failed: new data does not contain all variables (columns) required by model.'


class WrongExtension(BaseError):
    message = 'Cannot process file: wrong extension.'


class DelimiterError(BaseError):
    message = 'Parsing error: try changing delimiter.'


class NoColumnsError(BaseError):
    message = 'No columns in file.'
